/******************************************************************************
*
* project name:    TI-Chess
* file name:       cpc.c
* initial date:    28/03/2000
* author:          thomas.nussbaumer@gmx.net
* description:     compiler for chess puzzles
*
* $Id: cpc.c,v 1.3 2000/08/16 20:23:52 Thomas Nussbaumer Exp $
*
******************************************************************************/

/*---------------------------------------------------------------------------*/
/*                                                                           */
/* NOTE: This is just a quick and dirty hack to get it working ...           */
/*                                                                           */
/*---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


#define B_KING     -6          /* black figures      */
#define B_QUEEN    -5
#define B_KNIGHT   -4
#define B_BISHOP   -3
#define B_ROOK     -2
#define B_PAWN     -1
#define EMPTY       0          /* empty field        */
#define W_PAWN      1          /* white figures      */
#define W_ROOK      2
#define W_BISHOP    3
#define W_KNIGHT    4
#define W_QUEEN     5
#define W_KING      6

#define SIZE_TEXT      22
#define SIZE_COPYRIGHT 32

//#define SIZE_TEXT      20
//#define SIZE_COPYRIGHT 30
#define MAX_PROBLEMS   500

typedef struct {
   char          text1[SIZE_TEXT];
   char          text2[SIZE_TEXT];
   char          text3[SIZE_TEXT];
   char          copyright[SIZE_COPYRIGHT];
   unsigned char board[32];
   unsigned char unused;
   unsigned char side;
   unsigned char flgs_hi;
   unsigned char flgs_lo;
} problem_t;


problem_t problems[MAX_PROBLEMS];


FILE* infile    = 0;
FILE* outfile89 = 0;
FILE* outfile92 = 0;

char  actline[1024];

int counter = 0;
char board[8][8];
int  w_flags[8];
int  b_flags[8];
char text1[SIZE_TEXT];
char text2[SIZE_TEXT];
char text3[SIZE_TEXT];
char copyright[SIZE_COPYRIGHT];
int  side;
int  end_of_positions = 0;

long linecounter = 0L;

void Usage(char *programName) {
	fprintf(stderr,"usage: %s\nNOTE: file problems.txt will be converted\ninto ticpro.89s and ticpro.9xs\n\n",programName);
}


char* skiplines() {
    char* s = 0;
	int i;
	int l;
	do {
        linecounter++;
		s = fgets(actline,1023,infile);
        if (s == 0) return 0;

	    if (s[0] == '#') continue;

		l = strlen(actline);

        for (i=0;i<l;i++) {
	       if (!isspace(actline[i])) {
               //printf("%s\n",&actline[i]);
			   actline[l-1] = 0;
			   return &actline[i];
		   }
	    }
    }
	while(1);
    return s;
}


int convertBoard() {
    int i,j;

    for (i=0;i<8;i++) {
	    for (j=0;j<8;j++) {
		    switch(board[i][j]) {
		        case 'k': board[i][j] = W_KING;   break;
		        case 'q': board[i][j] = W_QUEEN;  break;
		        case 'n': board[i][j] = W_KNIGHT; break;
		        case 'b': board[i][j] = W_BISHOP; break;
		        case 'r': board[i][j] = W_ROOK;   break;
		        case 'p': board[i][j] = W_PAWN;   break;
		        case 'K': board[i][j] = B_KING;   break;
		        case 'Q': board[i][j] = B_QUEEN;  break;
		        case 'N': board[i][j] = B_KNIGHT; break;
		        case 'B': board[i][j] = B_BISHOP; break;
		        case 'R': board[i][j] = B_ROOK;   break;
		        case 'P': board[i][j] = B_PAWN;   break;
                case '_': board[i][j] = EMPTY;    break;
                case ' ': board[i][j] = EMPTY;    break;
				default: return 0;
		    }
        }
    }
	return 1;
}

int readPosition() {
   char* s;
   int i;

   if (!(s = skiplines())) {
       end_of_positions = 1;
	   return 0;
   }
   memset(text1,0,SIZE_TEXT);

   if (s[0] == '$') {
	   strcpy(text1," ");
   }
   else {
	   strncpy(text1,s,SIZE_TEXT);
   }

   if (!(s = skiplines())) {
       printf("ERROR in line %lu: cannot read second text line\n",linecounter);
	   return 0;
   }
   memset(text2,0,SIZE_TEXT);
   if (s[0] == '$') {
	   strcpy(text2," ");
   }
   else {
	   strncpy(text2,s,SIZE_TEXT);
   }

   if (!(s = skiplines())) {
       printf("ERROR in line %lu: cannot read third text line\n",linecounter);
	   return 0;
   }
   memset(text3,0,SIZE_TEXT);
   if (s[0] == '$') {
	   strcpy(text3," ");
   }
   else {
	   strncpy(text3,s,SIZE_TEXT);
   }

   if (!(s = skiplines())) {
       printf("ERROR in line %lu: cannot read copyright text line\n",linecounter);
	   return 0;
   }
   memset(copyright,0,SIZE_COPYRIGHT);
   if (s[0] == '$') {
	   strcpy(copyright," ");
   }
   else {
	   strncpy(copyright,s,SIZE_COPYRIGHT);
   }

   for (i=0;i<10;i++) {
	  if (!(s = skiplines())) {
           printf("ERROR in line %lu: cannot read board\n",linecounter);
		   return 0;
	  }

      if (i==0 || i == 9) {
		  if (strncmp(s,"abcdefgh",8) != 0) {
              printf("ERROR in line %lu: missing board marker \"abcdefgh\"\n",linecounter);
			  return 0;
		  }
		  else continue;
      }
      if (sscanf(&s[1],"%c%c%c%c%c%c%c%c",
	         &(board[i-1][0]),
	         &(board[i-1][1]),
	         &(board[i-1][2]),
	         &(board[i-1][3]),
	         &(board[i-1][4]),
	         &(board[i-1][5]),
	         &(board[i-1][6]),
	         &(board[i-1][7])) != 8) 
	  {
          printf("ERROR in line %lu: invalid number of characters in board\n",linecounter);
		  return 0;
	  }
   }

   if (!(s = skiplines())) return 0;
   if (sscanf(s,"%d%d%d%d%d%d%d%d",
	         &w_flags[0],
	         &w_flags[1],
	         &w_flags[2],
	         &w_flags[3],
	         &w_flags[4],
	         &w_flags[5],
	         &w_flags[6],
	         &w_flags[7]) != 8) 
   {
       printf("ERROR in line %lu: invalid number of white flags\n",linecounter);
	   return 0;
   }

   for (i=0;i<8;i++) {
	   if (w_flags[i] != 0 && w_flags[i] != 1) {
          printf("ERROR in line %lu: flags can only be 0 or 1\n",linecounter);
	      return 0;
	   }
   }

   if (!(s = skiplines())) return 0;
   if (sscanf(s,"%d%d%d%d%d%d%d%d",
	         &b_flags[0],
	         &b_flags[1],
	         &b_flags[2],
	         &b_flags[3],
	         &b_flags[4],
	         &b_flags[5],
	         &b_flags[6],
	         &b_flags[7]) != 8) 
	  {
          printf("ERROR in line %lu: invalid number of black flags\n",linecounter);
		  return 0;
	  }

   for (i=0;i<8;i++) {
	   if (b_flags[i] != 0 && b_flags[i] != 1) {
          printf("ERROR in line %lu: flags can only be 0 or 1\n",linecounter);
	      return 0;
	   }
   }

   if (!(s = skiplines())) {
       printf("ERROR in line %lu: missing \"side to move\" flag\n",linecounter);
	   return 0;
   }
   if (sscanf(s,"%d",&side) != 1) {
       printf("ERROR in line %lu: missing \"side to move\" flag\n",linecounter);
	   return 0;
   }

   if (side != 0 && side != 1) {
       printf("ERROR in line %lu: \"side to move\" can only be 0 or 1\n",linecounter);
	   return 0;
   }

   if (!convertBoard()) {
       printf("ERROR ABOVE line %lu: invalid character in board setup\n",linecounter);
	   return 0;
   }

   return 1;
}



void fillProblem(int index) {
    int i,j;
	char hi,lo;
	unsigned char both;
    int count = 0;

    strcpy(problems[index].text1,text1);
    strcpy(problems[index].text2,text2);
    strcpy(problems[index].text3,text3);
    strcpy(problems[index].copyright,copyright);


	for (i=7;i>=0;i--) {
       for (j=0;j<8;j+=2) {
           hi = board[i][j];
		   lo = board[i][j+1];
           if (lo<0) lo += 13;
           if (hi<0) hi += 13;
           both = (hi & 0x0f) << 4;
           both |= (lo & 0x0f);
           problems[index].board[count] = both;
		   count++;
	   }
    }

    problems[index].unused = 0;
    problems[index].side   = (unsigned char)side;

    //#define A1_MOVE_MASK      0x1000
    //#define B1_MOVE_MASK      0x0001
    //#define C1_MOVE_MASK      0x0002
    //#define E1_MOVE_MASK      0x0400
    //#define F1_MOVE_MASK      0x0004
    //#define G1_MOVE_MASK      0x0008
    //#define H1_MOVE_MASK      0x2000
    //#define W_ROCHADE_MASK    0x0010

    //#define A8_MOVE_MASK      0x4000
    //#define B8_MOVE_MASK      0x0020
    //#define C8_MOVE_MASK      0x0040
    //#define E8_MOVE_MASK      0x0800
    //#define F8_MOVE_MASK      0x0080
    //#define G8_MOVE_MASK      0x0100
    //#define H8_MOVE_MASK      0x8000
    //#define B_ROCHADE_MASK    0x0200

    problems[index].flgs_hi =
	     b_flags[5] * 0x01 + b_flags[7] * 0x02 +   // G8 + B_ROCHADE
         w_flags[3] * 0x04 + b_flags[3] * 0x08 +   // E1 + E8
         w_flags[0] * 0x10 + w_flags[6] * 0x20 +   // A1 + H1
         b_flags[0] * 0x40 + b_flags[6] * 0x80;    // A8 + H8

    problems[index].flgs_lo =
         w_flags[1] * 0x01 + w_flags[2] * 0x02 +   // B1 + C1
         w_flags[4] * 0x04 + w_flags[5] * 0x08 +   // F1 + G1
         w_flags[7] * 0x10 + b_flags[1] * 0x20 +   // W_ROCHADE + B8
         b_flags[2] * 0x40 + b_flags[4] * 0x80;    // C8 + F8

}

typedef struct {
    char          signature[8]; // "**TI92P*" or "**TI89**"
    unsigned char fill1[2];     // 01 00
    char          folder[8];    // folder name
    char          desc[40];     //
    unsigned char fill2[6];     // 01 00 52 00 00 00
    char          name[8];      //
    unsigned char type[4];      // 0C 00 00 00
    char          size[4];      // complete file
    unsigned char fill3[6];     // A5 5A 00 00 00 00
    unsigned char datasize[2];  
}
header_t;

void write2Output(char* signature,FILE* fp) {
    header_t header;

    int size = sizeof(header_t)+sizeof(problem_t)*counter+3+4;
    int i,j;
    unsigned int   checksum;
    unsigned char* ptr;

	memset(&header,0,sizeof(header_t));
    strcpy(header.signature,signature);
	header.fill1[0] = 1;
	header.fill1[1] = 0;
    strcpy(header.folder,"main");
    header.fill2[0] = 0x01;
    header.fill2[1] = 0x00;
    header.fill2[2] = 0x52;
    header.fill2[3] = 0x00;
    header.fill2[4] = 0x00;
    header.fill2[5] = 0x00;
	strcpy(header.name,"ticpro");
	header.type[0] = 0x0c;  // 0c
	header.type[1] = 0x00;
	header.type[2] = 0x03;
	header.type[3] = 0x00;
	header.size[0] = size % 256;
	header.size[1] = size / 256;
	header.size[2] = 0;
	header.size[3] = 0;
	header.fill3[0] = 0xA5;
	header.fill3[1] = 0x5A;
	header.fill3[2] = 0x00;
	header.fill3[3] = 0x00;
	header.fill3[4] = 0x00;
	header.fill3[5] = 0x00;

	size -= sizeof(header_t);
	size -= 2;

	header.datasize[0] = size / 256;
	header.datasize[1] = size % 256;

	checksum = header.datasize[0];
	checksum += header.datasize[1];

	ptr = (unsigned char*)&header;
	for (i=0;i<sizeof(header_t);i++) {
	    fputc(*ptr,fp);
		ptr++;
    }

	checksum += counter%256;
	checksum += counter/256;

	fputc(0,fp);
	
	fputc(counter/256,fp);
	fputc(counter%256,fp);    

	for (i=0;i<counter;i++) {
		ptr = (unsigned char*)&problems[i];
		for (j=0;j<sizeof(problem_t);j++) {
  	        fputc(*ptr,fp);
			checksum+=*ptr;
		    ptr++;
	    }
    }

	fputc(0,fp);
	fputc(0x2D,fp);

	checksum+=0x2d;

	fputc(checksum%256,fp);
	fputc(checksum/256,fp);

}



int main(int argc,char *argv[]) {

	if (argc != 1) {
		Usage(argv[0]);
		return 1;
	}

    if (!(infile = fopen("problems.txt","r"))) {
		printf("ERROR: cannot open infile problems.txt\n");
        return 1;
    }

    if (!(outfile89 = fopen("ticpro.89s","wb"))) {
		printf("ERROR: cannot open outfile ticpro.89s\n");
        fclose(infile);
		return 1;
    }

    if (!(outfile92 = fopen("ticpro.9xs","wb"))) {
		printf("ERROR: cannot open outfile ticpro.9xs\n");
        fclose(infile);
        fclose(outfile92);
		return 1;
    }

	do {
        if (!readPosition() && end_of_positions == 0) {
		    printf("ERROR: cannot read position %d\n",counter);
            break;
        }

	    if (end_of_positions == 0) {
            fillProblem(counter);
			counter++;
		    printf("position %d read ...\n",counter);
        }
    }
	while (end_of_positions == 0);

    if (counter == 0) {
	    printf("ERROR: no positions read.\n");
    }

    else if (end_of_positions == 1) {
	    printf("%d positions read.\nexcuting save now ...\n",counter);

		write2Output("**TI89**",outfile89);
		write2Output("**TI92P*",outfile92);

    }


	fclose(infile);
	fclose(outfile89);
	fclose(outfile92);

	//printf("sizeof(header)=%d sizeof(problem)=%d\n",sizeof(header_t),sizeof(problem_t));

	return 0;
}

//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: cpc.c,v $
// Revision 1.3  2000/08/16 20:23:52  Thomas Nussbaumer
// variable datasize in structure header_t changed to unsigned char
// for correctness
//
// Revision 1.2  2000/08/12 15:33:18  Thomas Nussbaumer
// substitution keywords added
//
//
